# 
# Copyright (C) 2003 Carl Johan Lejdfors
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# Please send all bugreports to: Carl Johan Lejdfors <calle.lejdfors@cs.lth.se>
# 

#import sys
#sys.path.append("../../2003/pycg")

import tempfile
import sys
import os
import re
import string
from copy import *
from Cg import *
from OpenGL.GL.ARB.multitexture import *



## default context stuff
_cgContext = cgCreateContext()
cgSetAutoCompile(_cgContext, CG_COMPILE_MANUAL)

_cgStateParameter = {'modelview': (CG_GL_MODELVIEW_MATRIX,
                                   CG_GL_MATRIX_IDENTITY),
                     'modelviewi': (CG_GL_MODELVIEW_MATRIX,
                                    CG_GL_MATRIX_INVERSE),
                     'modelviewt': (CG_GL_MODELVIEW_MATRIX,
                                    CG_GL_MATRIX_TRANSPOSE),
                     'modelviewit': (CG_GL_MODELVIEW_MATRIX,
                                     CG_GL_MATRIX_INVERSE_TRANSPOSE),
                     'projection': (CG_GL_PROJECTION_MATRIX,
                                    CG_GL_MATRIX_IDENTITY),
                     'projectioni': (CG_GL_PROJECTION_MATRIX,
                                     CG_GL_MATRIX_INVERSE),
                     'projectiont': (CG_GL_PROJECTION_MATRIX,
                                     CG_GL_MATRIX_TRANSPOSE),
                     'projectionit': (CG_GL_PROJECTION_MATRIX,
                                      CG_GL_MATRIX_INVERSE_TRANSPOSE),
                     'texture': (CG_GL_TEXTURE_MATRIX,
                                 CG_GL_MATRIX_IDENTITY),
                     'texturei': (CG_GL_TEXTURE_MATRIX,
                                  CG_GL_MATRIX_INVERSE),
                     'texturet': (CG_GL_TEXTURE_MATRIX,
                                  CG_GL_MATRIX_TRANSPOSE),
                     'textureit': (CG_GL_TEXTURE_MATRIX,
                                   CG_GL_MATRIX_INVERSE_TRANSPOSE),
                     'modelviewproj': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                       CG_GL_MATRIX_IDENTITY),
                     'modelviewproji': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                        CG_GL_MATRIX_INVERSE),
                     'modelviewprojt': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                        CG_GL_MATRIX_TRANSPOSE),
                     'modelviewprojit': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                         CG_GL_MATRIX_INVERSE_TRANSPOSE),
                     'worldviewproj': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                       CG_GL_MATRIX_IDENTITY),
                     'worldviewproji': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                        CG_GL_MATRIX_INVERSE),
                     'worldviewprojt': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                        CG_GL_MATRIX_TRANSPOSE),
                     'worldviewprojit': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                         CG_GL_MATRIX_INVERSE_TRANSPOSE),
                     'world': (CG_GL_MODELVIEW_MATRIX,
                               CG_GL_MATRIX_IDENTITY),
                     'worldit': (CG_GL_MODELVIEW_MATRIX,
                                 CG_GL_MATRIX_INVERSE_TRANSPOSE),
                     'viewit': (CG_GL_MODELVIEW_MATRIX,
                                CG_GL_MATRIX_INVERSE_TRANSPOSE)
                     }

vp20 = CG_PROFILE_VP20
fp20 = CG_PROFILE_FP20
vp30 = CG_PROFILE_VP30
fp30 = CG_PROFILE_FP30
arbvp1 = CG_PROFILE_ARBVP1
arbfp1 = CG_PROFILE_ARBFP1


class Cg:

    def __init__(self, string):
        self.string = string

    def __call__(self,**args):
        return CgFactory(self.string)(**args)

    def __getattr__(self, attr):
        return CgFactory(self.string,entry=attr)

    def __str__(self):
        return 'Cg("""'+self.string+'""")'

    def __nonzero__(self):
        return True
    
class CgFactory:

    def __init__(self, string, entry = 'main'):
        self.string = string
        self.entry = entry

        self.cgTextures = []
        self.cgClientStates = []
        self.parameters = []
        self.linkList = []

        self.cache = {}

    def __call__(self,
                 target = None,
                 args = None,
                 **kwords):
        self.target = target
        self.args = None
        self.kwords = kwords
        return self

    def clearCache(self):
        self.cache = {}

    def bindVertex(self, ctx):
        if not self.target:
            self.target = cgGLGetLatestProfile(CG_GL_VERTEX)
        return self._bind(ctx)

    def bindFragment(self, ctx):
        if not self.target:
            self.target = cgGLGetLatestProfile(CG_GL_FRAGMENT)
        return self._bind(ctx)

    def _bind(self, ctx):
        self.compile()
        p = self.link(ctx)
        if p:
            p.bind()
            p.set_parameters(ctx)
            return p

    def bind(self):
        assert(cgIsProgramCompiled(self.prog))
        cgGLEnableProfile(self.target)
        #cgGLLoadProgram(self.prog)
        cgGLBindProgram(self.prog)
            
    def reset(self):
        map(cgGLDisableTextureParameter, self.cgTextures)
        map(cgGLDisableClientState, self.cgClientStates)
        self.cgClientStates = []
        self.cgTextures = []
        cgGLDisableProfile(self.target)
        

    def compile(self):
        if hasattr(self,'prog'):
            return

        self.prog = cgCreateProgram(_cgContext,
                                    CG_SOURCE,
                                    self.string,
                                    self.target,
                                    self.entry,
                                    self.args)

        if not self.prog:
            print self.string
            raise RuntimeError( 'Unable to create program: %s' % 
                                cgGetErrorString(cgGetError()) )

        ##iterate over all parameters and build the link-table
        param = cgGetFirstParameter(self.prog, CG_PROGRAM)
        while param:
            t = cgGetParameterNamedType(param)
            name = cgGetParameterName(param)
            if t >= 32768:
                type = cgGetTypeString(t)
                if cgIsInterfaceType(t):
                    self.linkList.append((name,type))
            elif cgGetParameterVariability(param) == CG_UNIFORM and \
                     cgGetParameterDirection(param) in [CG_IN, CG_INOUT, CG_ERROR]:
                ## need to have CG_ERROR as well since the direction of an
                ## array is apparently not defined. The other two are needed for
                ## some stupid profiles which incorrectly returns the variability
                ## of some uniforms
                ## /Calle
                self.add_parameter(name)
                
            param = cgGetNextParameter(param)


        if not self.linkList:
            ##trivial shader compile now
            cgCompileProgram(self.prog)
            cgGLLoadProgram(self.prog)

    def computeCacheEntry(self, fragments):
        #return "s" + str(hash(self)) + "_" + \
        return string.join(map(lambda f:str(f.type),fragments),'_')

    def link(self, context):
        if not self.linkList:
            #print "Trivial shader, no linking required!"
            return self

        #try:
        fs = map(lambda (n,t): getattr(context,n), self.linkList)
        
        cacheEntry = self.computeCacheEntry(fs)
        if self.cache.has_key(cacheEntry):
            #print "Using cached value", cacheEntry
            return self.cache[cacheEntry]
        
        shader = CgShader(self.string,
                          self.target,
                          self.entry,
                          self.args,
                          self.kwords)
        
        ## First iterate over all program parameters
        ## link all fragments and build the external parameter list
        for (name, type), fragment in zip(self.linkList,fs):
            if fragment.interface == type:
                shader.link_fragment(name, fragment)
            else:
                print "Found %s with interface %s but wanted type %s" \
                      % (name, fragment.interface, type)


        for name,param in self.parameters:
            shader.add_parameter(name,param)

        ## Now the program should be compileable
        shader.compile()

        #print "caching shader as", cacheEntry
        self.cache[cacheEntry] = shader

        return shader

        #except AttributeError, e:
        #    print "Unable to compile program since:", e
        

    def add_parameter(self, name, param = None): ## -> None
        ##print "adding parameter", name
        self.parameters.append((name,param))

    def set_parameters(self, context):
        #print "binding %s parameters!" % len(self.parameters)
        for pname,param in self.parameters:
            if not param:
                param = cgGetNamedParameter(self.prog, pname)

            self.set_parameter(context, pname, param)

    def set_parameter(self, context, pname, param):
        value = None
        
        try:
            value = self.kwords[pname]
        except:
            pass

        try:
            if value == None:
                value = getattr(context,pname)
        except:
            pass

        try:
            if value == None:
                matrix, modifier = _cgStateParameter[pname.lower()]
                cgGLSetStateMatrixParameter(param, matrix, modifier)
                return 
        except:
            pass

                

        ## first use compile-time parameters
        #         if self.kwords.has_key(pname):
        #             #cgSetParameterVariability(param, CG_LITERAL)
        #             value = self.kwords[pname]
        
        #         ## next, look in the resolver
        #         elif hasattr(context, pname):
        #             value = getattr(context, pname)
        
        #         ## else, check the state matrix parameters
        #         elif _cgStateParameter.has_key(pname.lower()):
        #             matrix, modifier = _cgStateParameter[pname.lower()]
        #             cgGLSetStateMatrixParameter(param, matrix, modifier)
        #             return 
        
        #         else:
        if value == None:
            print "Unable to find parameter", pname
            return

        self._set_parameter(param, value)
        

    def _set_parameter(self, param, value):
        cgType = cgGetParameterType(param)
        
        if (cgGetParameterVariability(param) == CG_VARYING and
            cgType != CG_STRUCT):  
            cgGLEnableClientState(param)
            value = asarray(value, Float)
            cgGLSetParameterPointer(param, value.shape[-1],
                                    GL_DOUBLE, 0, value),
            self.cgClientStates.append(param)
            
        elif cgType == CG_FLOAT:
            cgGLSetParameter1f(param, value)

        elif cgType == CG_FLOAT2:
            cgGLSetParameter2f(param, value[0], value[1])

        elif cgType == CG_FLOAT3:
            cgGLSetParameter3f(param, value[0],
                               value[1], value[2])

        elif cgType == CG_FLOAT4:
            cgGLSetParameter4f(param,
                               value[0], value[1],
                               value[2], value[3])

        elif cgType in [CG_FLOAT1x1, CG_FLOAT1x2, CG_FLOAT1x3,
                        CG_FLOAT1x4, CG_FLOAT2x1, CG_FLOAT2x2,
                        CG_FLOAT2x3, CG_FLOAT2x4, CG_FLOAT3x1,
                        CG_FLOAT3x2, CG_FLOAT3x3, CG_FLOAT3x4,
                        CG_FLOAT4x1, CG_FLOAT4x2, CG_FLOAT4x3,
                        CG_FLOAT4x4]:
            cgGLSetMatrixParameterdc(param, value)

        elif cgType in [CG_SAMPLER1D, CG_SAMPLER2D,
                        CG_SAMPLER3D, CG_SAMPLERCUBE, CG_SAMPLERRECT]:
            ## UGLY, should not rely on internal state information
            texNo = value.textureId()
            cgGLSetTextureParameter(param, texNo)
            cgGLEnableTextureParameter(param)

            texUnit = cgGetParameterResource(param) - CG_TEXUNIT0
            if texUnit:
                glActiveTextureARB(GL_TEXTURE0_ARB+texUnit)

            value.bind(0)
            self.cgTextures.append(param)

        elif cgType == CG_ARRAY:
            for i in range(cgGetArraySize(param,0)):
                aparam = cgGetArrayParameter(param, i)
                self._set_parameter(aparam, value[i])

        elif cgType == CG_STRUCT:
            sparam = cgGetFirstStructParameter(param)
            while sparam:
                name = string.split(cgGetParameterName(sparam),'.')[-1]
                val = None
                if hasattr(value, name):
                    val = getattr(value, name)
                elif hasattr(value, name+'s'):
                    val = getattr(value, name+'s')
                if val:
                    self.setParameter(sparam,val)
                sparam = cgGetNextParameter(sparam)

        else:            
            raise RuntimeError('Unknown type %s' % cgGetTypeString(type))



class CgFragment:

    def __init__(self, interface, type, string, **kwords):
        self.interface = interface
        self.type = type
        self._string = string
        self.kwords = kwords

    def string(self):
        return "struct %s : %s { %s };" % (self.type,self.interface,self._string)



class CgShader(CgFactory):

    def __init__(self, string, target, entry, args, kwords):
        CgFactory.__init__(self, string, entry)
        self(target=target, args = args, **kwords)

        self.fragments = []
        self._cgFragmentParameters = []
        
    def __del__(self):
        ## print "Cleaning up %s parameters" % len(self._cgFragmentParameters)
        ## clean up all created fragment parameters
        map(cgDestroyParameter, self._cgFragmentParameters)
        
    def link_fragment(self, name, fragment): ## -> None
        #print "linking fragment", fragment.type, "under name", name

        programString = self.string
        mainRE = re.compile( "(\w+) main\(" )
        match = mainRE.search(programString)

        if match:
            newProgramString = programString[0:match.start()]
            newProgramString += fragment.string() + "\n\n"
            newProgramString += programString[match.start():]

            self.string = newProgramString
            
        self.fragments.append((name, fragment))


    def compile(self):
        #print "Compiling program"
        #print self.string

        ##Same as for shader factory
        if hasattr(self,'prog'):
            return

        self.prog = cgCreateProgram(_cgContext,
                                    CG_SOURCE,
                                    self.string,
                                    self.target,
                                    self.entry,
                                    self.args)

        if not self.prog:
            if self.cgc:
                fd,n = tempfile.mkstemp('.cg')
                f = os.fdopen(fd,'w')
                f.write(self.string)
                f.flush()
                
                profileDict = { arbvp1 : 'arbvp1', vp20 : 'vp20', \
                                vp30   : 'vp30', arbfp1 : 'arbfp1', \
                                fp20   : 'fp20', fp30   : 'fp30' }
                
                stdin,stdout = os.popen4(self.cgc + \
                                         ' -profile ' + profileDict[self.target] + \
                                         ' -entry ' + self.entry + ' ' + n, 't') 
                errors = stdout.read()
                
                m = re.findall("(.*)\(([0123456789]+)\) : (.*)", errors)
                
                f.close()
                os.unlink(n)
                
                if m:
                    self.prog = None
                    raise RuntimeError('Unable to compile program:\n %s' %
                                       string.join(map(lambda (f,ln,e): self.entry + \
                                                   '(%s): %s' % (ln,e), m),'\n'))
            print self.string
            raise RuntimeError( 'Unable to create program: %s' % 
                                cgGetErrorString(cgGetError()) )

        ## end same

        ## create and connect parameters for all fragments
        ## add new external parameters
        for pname, fragment in self.fragments:
            #print "linking fragment:", pname
            param = cgCreateParameter(_cgContext,
                                      cgGetNamedUserType(self.prog, fragment.type))
            cgConnectParameter(param, cgGetNamedParameter(self.prog, pname))
            self._cgFragmentParameters.append(param)
            
            sparam = cgGetFirstStructParameter(param)
            while sparam:
                pname = string.split(cgGetParameterName(sparam), '.')[-1]
                p = cgCreateParameter(_cgContext, cgGetParameterType(sparam))
                cgConnectParameter(p, sparam)
                try:
                    value = getattr(fragment,pname)
                    self._set_parameter(p, value)
                    self._cgFragmentParameters.append(p)
                except AttributeError:
                    print "Unable to resolve fragment parameter, adding as external!"
                    self.add_parameter(pname,p)
                sparam = cgGetNextParameter(sparam)

            
        ## now compile and load the program
        cgCompileProgram(self.prog)
        assert(cgIsProgramCompiled(self.prog))

        cgGLLoadProgram(self.prog)
        
        
        #print cgGetProgramString(self.prog, CG_COMPILED_PROGRAM)

    def bind(self):
        assert(cgIsProgramCompiled(self.prog))
        cgGLEnableProfile(self.target)
        #cgGLLoadProgram(self.prog)
        cgGLBindProgram(self.prog)
            
    def reset(self):
        cgGLDisableProfile(self.target)
        pass


             

        
## Test if there is an available cgc in
## the Cg installation path or the system path.
##
## /Calle
if os.name == 'nt':
    cgcExecutable = 'cgc.exe'
else:
    cgcExecutable = 'cgc'
cgcPath = os.path.join(sys.prefix,'lib','site-packages','Cg',cgcExecutable)
try:
    h,n = tempfile.mkstemp('.cg')
    i,o,e = os.popen3(cgcPath + ' -help')

    hasCGC = None
    if o.read():
        hasCGC = cgcPath

    else: ## try the standard path
        i,o,e = os.popen3(cgcExecutable + ' -help')
                
        if o.read():
            hasCGC = cgcExecutable

    CgFactory.cgc = hasCGC

finally:
    os.close(h)
    os.unlink(n)
