# 
# Copyright (C) 2003 Carl Johan Lejdfors
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# Please send all bugreports to: Carl Johan Lejdfors <calle.lejdfors@cs.lth.se>
# 

import FX 
import FX.util
import types

import FX.cg_shader

class BasicResolver:

    def __init__(self, nameMap, *contexts):
        self.nameMap = nameMap
        self.contexts = list(contexts)
        #dict = {}
        #contexts = list(contexts)
        #contexts.reverse()
        #for ctx in contexts:
        #    dict.update(ctx.__dict__)
        #self.dict = dict

    def __getattr__(self, attr):
        #print "finding", attr
        try:
            attr = self.nameMap[attr]
        except KeyError:
            pass

        #return self.dict[attr]
        
        for ctx in self.contexts:
            try:
                return getattr(ctx, attr)
            except:
                pass
        #try:
        #    attr = self.nameMap[attr]
        #    for ctx in self.contexts:
        #        try: #if hasattr(ctx, attr):
        #            return getattr(ctx, attr)
        #        except:
        #            pass
        #except KeyError:
        #    pass
        
        raise AttributeError("Unable to resolve '%s'" % attr)

class Effect:
    """ The top-level effect class. To load an effect from a module
        located in the path use:

          e = Effect('MyEffect')

        This effect can the be used as:

          for p in e.passes():
              p(mesh)
              renderMesh(mesh)
    
        If you subclass this class you need to provide the 'name' and
        'passes' attributes. The moduleName constructor argument
        should not be used as this implies loading the effect from the
        corresponding file. """
        
         
    def __init__(self,
                 moduleName = None,
                 nameMap = {},
                 resolverClass = BasicResolver,
                 **kwords):
        ## update my dictionary and set the name of the effect
        ## if we're using a module-style effect
        self.theModule = None
        if moduleName:
            self.name = moduleName
            self.init = []
            #self.__dict__.update(__import__(moduleName).__dict__)
            self.theModule = __import__(moduleName)

            self.technique = self.theModule.technique
            try:
                self.init = self.theModule.initialize
            except:
                self.init = []
                
            #if type(self.technique[0]) == types.ListType:
            #    self._passes = self.technique[0]
            #else:
            #    self._passes = self.technique
        
        ##allow constructor arguments override of effect parameters
        #self.__dict__.update(kwords)

        self.tech = None
        self._resolverClass = resolverClass
        self._nameMap = nameMap

    #def __getattr__(self, attr):
    #    try:
    #        return self.__dict__[attr]
    #    except KeyError:
    #        try:
    #            return getattr(self.theModule, attr)
    #        except AttributeError:
    #            raise AttributeError("Could not find '%s' in effect!" % attr)

    def initialize(self):
        for p in self.init:
            FX.State.Begin()

            if p.use(FX.State, lambda *ctxs: self._resolverClass(self._nameMap,self,*ctxs)):
                print "Can't use render type steps in init!"

            FX.State.Reset()

    def setTechnique(self, i):
        self.tech = i
        
    def passes(self, *contexts):
        contexts = list(contexts)

        if self.tech == None:
            ps = self.technique
        else: 
            ps = self.technique[self.tech]

        contexts += [self]
        
        if not self.theModule == None :
            contexts += [self.theModule] 

        for p in ps:
            FX.State.Begin()

            if p.use(FX.State, lambda *ctxs: self._resolverClass(self._nameMap,*(list(ctxs)+contexts))):
                yield p

            FX.State.Reset()

class Render:
    def __init__(self, **kwords):
        self.kwords = kwords

    def __call__(self, *ctxs):
        self.state.resolver = self.resolverCtor(*ctxs)
        
        for s,v in self.kwords.items():
            #print "Setting state %s with %s" % (s,v)
            
            FX.util.marshalFX(self.state,s,v)
            try:

                #FX.util.marshalFX(self.state,s,v)
                pass
             
            except AttributeError, e:
                print "State::ERROR Unknown state '%s': '%s'" % (s,e)

            except AssertionError, e:
                print "An assertion was raised when calling '%s' with '%s': '%s'" % (s,v,e)

            except NotImplementedError, e:
                print "State '%s' is not implemented" % key

            except RuntimeError, e:
                print "Setting state '%s' with '%s' failed with: '%s'" % (s,v,e)

            except Exception, e:
                print "Setting state '%s' with '%s' terminated with unknown exception: %s" % (s,v,e)

        self.state.resolver = None

            
    def use(self, state, resolverCtor):
        self.state = state
        self.resolverCtor = resolverCtor
        return True


class ImageProcessing(Render):
    __showFS = FX.cg_shader.Cg("""
    float4 main(float2 texcoords : TEXCOORD,
                uniform samplerRECT Source) : COLOR
    {
       return texRECT(Source, texcoords);
    }
    """)
    
    def __init__(self, sources, **kwords):
        defaults = { 'Lighting' : False,
                     'ZWriteEnable' : False, 
                     'FragmentShader' : ImageProcessing.__showFS(**kwords) }
        defaults.update(kwords)

        if not isinstance(sources, list):
            sources = [sources]
            
        self.sizes = map(lambda x: x.size(), sources)
        Render.__init__(self,**defaults)

    def use(self, state, *args):
        Render.use(self,state,*args)
        self()
        state.draw2DPlane(self.sizes)
        return False

# class Group(Render):
#     def __init__(self,
#                  Target = None,
#                  Clear = None,
#                  **kwords):
#         Render.__init__(self,**kwords)
#         self.target = Target
#         self.clear = Clear

#     def use(self, state):
#         if self.target:
#             self.target.Activate()
#             state.b

from OpenGL.GL import *

class Target:
    def __init__(self, target, **kwords):
        self.target = target
        self.kwords = kwords

    def use(self, state, *args):
        mv = glGetFloat(GL_MODELVIEW_MATRIX)
        proj = glGetFloat(GL_PROJECTION_MATRIX)
        
        self.target.activate()
        state.Begin()

        glMatrixMode(GL_PROJECTION)
        glLoadMatrixd(proj)

        glMatrixMode(GL_MODELVIEW)
        glLoadMatrixd(mv)
        
        for s,v in self.kwords.items():
            #glClearColor(1,1,1,0)
            FX.util.marshalFX(state,s,v)

        return False
                       
class EndTarget(Target):
    def use(self, state, resolverCtor):
        #self.target.copyToTexture()
        state.Reset()
        self.target.deactivate()

        return False


# class BindTexture(Render):
#     def __init__(self, texture):
#         self.texture = texture
        
#     def use(self, state):
#         self.texture.bind()
