# -*- coding: iso-8859-1 -*-
# 
# Copyright (C) 2003 Carl Johan Lejdfors
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# Please send all bugreports to: Carl Johan Lejdfors <calle.lejdfors@cs.lth.se>
# 


from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.GL.ARB.multitexture import *
from OpenGL.GL.NV.texture_rectangle import *
from FX.constants import *
from FX.util import *
from Numeric import asarray, Float

## Alla states som r mjliga att stta i .fx finns dokumenterade
## p:
##
##   http://msdn.microsoft.com/library/default.asp?url=/library/en-us/directx9_c/directx/graphics/reference/EffectFileReference/EffectFileFormat/States.asp
##
## /Calle <calle.lejdfors@cs.lth.se>
##

class OpenGLRenderState:

    def __init__(self):
        self.alphaFunc = GL_ALWAYS
        self.alphaRef = 0.0
        self.destBlend = GL_ZERO
        self.sourceBlend = GL_ONE
        self.polygonFace = GL_FRONT_AND_BACK
        self.polygonMode = GL_FILL
        self.stencilFail = GL_KEEP
        self.zFail = GL_KEEP
        self.zPass = GL_KEEP
        self.stencilFunc = GL_ALWAYS
        self.stencilRef = 0
        self.stencilMask = ~0
        self.clipPlanes = []

    def _uglyINIT(self):
        """ This must be called *after* the OpenGL context has been created in
            order to work around a multitexture problem on the Windows platform."""
        self.supportsMultitexture = glInitMultiTexARB()
        self.noTexUnits = glGetInteger(GL_MAX_TEXTURE_UNITS_ARB) 

        self.supportsRECT = glInitTextureRectangleNV()

       
    def Begin(self):
        glPushAttrib( GL_ALL_ATTRIB_BITS )

    def Reset(self):
        glPopAttrib()


    def __glenable(self, enum, val):
        """ Handles all kinds of state settings of the form 'glEnable(GL_XXX)'. """
        if val:
            glEnable(enum)
        else:
            glDisable(enum)

    def AlphaBlendEnable(self, val):
        self.__glenable(GL_BLEND,val)

    def AlphaFunc(self, val):
        self.alphaFunc = val
        glAlphaFunc(self.alphaFunc,self.alphaRef)

    def AlphaRef(self, val):
        self.alphaRef = val
        glAlphaFunc(self.alphaFunc,self.alphaRef)

    def AlphaTestEnable(self, val):
        self.__glenable(GL_ALPHA_TEST,val)

    def BlendOp(self, val):
        ##!!FIXME!! is this correct?
        if val == ADD:
            glBlendFunc(GL_SRC_ALPHA, GL_ONE)
        elif val == SUBTRACT:
            ## glBlendFunc(GL_SRC_ALPHA,
            raise NotImplementedError()
        elif val == REVSUBTRACT:
            raise NotImplementedError()
        elif val == MIN:
            raise NotImplementedError()
        elif val == MAX:
            raise NotImplementedError()
        else:
            raise RuntimeError("Unknown value for BlendOp '%s'" % val)

    def ColorWriteEnable(self, val):
        if val:
            glColorMask(255,255,255,255)
        else:
            glColorMask(0,0,0,0)

    def DepthBias(self, val):
        #raise NotImplementedError()
        glPixelTransferf(GL_DEPTH_BIAS,val)
    
    def DestBlend(self, val):
        assert val in [ZERO, ONE,
                       SRCCOLOR, INVSRCCOLOR,
                       SRCALPHA,INVSRCALPHA,
                       DESTALPHA, INVDESTALPHA]
        self.destBlend = val
        glBlendFunc(self.sourceBlend, self.destBlend)

    def DitherEnable(self, val):
        self.__glenable(GL_DITHER,val)

    def FillMode(self, val):
        assert val in [POINT,WIREFRAME,SOLID]
        self.polygonMode = val
        glPolygonMode(self.polygonFace, self.polygonMode)

    def LastPixel(self, val):
        raise NotImplementedError()

    def ShadeMode(self, val):
        assert val in [FLAT,GOURAUD,PHONG]
        glShadeModel(val)

    def SrcBlend(self, val):
        assert val in [ZERO, ONE,
                       DESTCOLOR, INVDESTCOLOR,
                       SRCALPHA,INVSRCALPHA,
                       DESTALPHA, INVDESTALPHA,
                       SRCCOLOR, INVSRCCOLOR,
                       SRCALPHASAT]
        self.sourceBlend = val
        glBlendFunc(self.sourceBlend, self.destBlend)

    def StencilEnable(self, val):
        if val:
            glEnable(GL_STENCIL_TEST)
        else:
            glDisable(GL_STENCIL_TEST)

    def StencilFail(self, val):
        assert val in [KEEP,ZERO,REPLACE,
                       INCRSAT,DECRSAT,
                       INVERT,INCR,DECR,TWOSIDED]
        self.stencilFail = val
        glStencilOp(self.stencilFail,
                    self.zFail,
                    self.zPass)

    def StencilFunc(self, val):
        assert val in [GL_NEVER, GL_LESS, GL_LEQUAL, GL_GREATER,
                       GL_GEQUAL, GL_EQUAL, GL_NOTEQUAL, GL_ALWAYS ]
        self.stencilFunc = val
        glStencilFunc(self.stencilFunc,
                      self.stencilRef,
                      self.stencilMask)

    def StencilMask(self, val):
        self.stencilMask = val
        glStencilFunc(self.stencilFunc,
                      self.stencilRef,
                      self.stencilMask)

    def StencilPass(self, val):
        assert val in [KEEP,ZERO,REPLACE,
                       INCRSAT,DECRSAT,
                       INVERT,INCR,DECR,TWOSIDED]
        self.zPass = val
        glStencilOp(self.stencilFail,
                    self.zFail,
                    self.zPass)

    def StencilRef(self, val):
        self.stencilRef = val
        glStencilFunc(self.stencilFunc,
                      self.stencilRef,
                      self.stencilMask)

    def StencilWriteMask(self, val):
        glStencilMask(val)

    def StencilZFail(self, val):
        assert val in [KEEP,ZERO,REPLACE,
                       INCRSAT,DECRSAT,
                       INVERT,INCR,DECR,TWOSIDED]
        self.zFail = val
        glStencilOp(self.stencilFail,
                    self.zFail,
                    self.zPass)

    def TextureFactor(self, val):
        raise NotImplementedError()
    
    def ZEnable(self, val):
        self.__glenable(GL_DEPTH_TEST,val)

    def ZFunc(self, val):
        assert val in [NEVER, LESS, EQUAL, LESSEQUAL,
                       GREATER, NOTEQUAL, GREATEREQUAL, ALWAYS]
        glDepthFunc(val)

    def ZWriteEnable(self, val):
        if val:
            glDepthMask(~0)
        else:
            glDepthMask(GL_FALSE)

 ## Vertex pipe-line states
    def Ambient(self, val):
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, val)

    def AmbientMaterialSource(self, val):
        raise NotImplementedError()

    def Clipping(self, val):
        raise NotImplementedError()

    def ClipPlaneEnable(self, val):
        i = val - GL_MAX_CLIP_PLANES
        glClipPlane( val, self.clipPlanes[i] )

    def Color(self, val):
        if len(val) == 3:
            glColor3f( *val )
        elif len(val) == 4:
            glColor4f( *val )
        else:
            raise RuntimeError("Color '%s' not RGB(A)")

    def ColorMaterial(self, val):
        if val:
            glEnable(GL_COLOR_MATERIAL)
            glColorMaterial(GL_FRONT_AND_BACK, val)
        else:
            glDisable(GL_COLOR_MATERIAL)
        
    def ColorVertex(self, val):
        raise NotImplementedError()

    def CullMode(self, val):
        if val:
            glEnable(GL_CULL_FACE)
            glFrontFace(val)
        else:
            glDisable(GL_CULL_FACE)

    def CullFace(self, val):
        if val:
            glEnable(GL_CULL_FACE)
            glCullFace(val)
        else:
            glDisable(GL_CULL_FACE)

    def DiffuseMaterialSource(self, val):
        raise NotImplementedError()

    def EmissiveMaterialSource(self, val):
        raise NotImplementedError()
    
    def FogColor(self, val):
        glFogfv(GL_FOG_COLOR, val)

    def FogDensity(self, val):
        glFogfv(GL_FOG_DENSITY, val)

    def FogEnable(self, val):
        self.__glenable(GL_FOG,val)

    def FogEnd(self, val):
        glFogf(GL_FOG_END, val)

    def FogState(self, val):
        glFogf(GL_FOG_START, val)

    def FogTableMode(self, val):
        if not val:
            self.FogEnable(False)
        else:
            assert val in [LINEAR, EXP, EXP2]
            glFogf(GL_FOG_MODE,val)

    def FogVertexMode(self, val):
        raise NotImplementedError()

    def IndexedVertexBlendEnable(self, val):
        raise NotImplementedError()

    def Lighting(self, val):
        self.__glenable(GL_LIGHTING, val)

    def Light(self, i, val):
        glLight = GL_LIGHT0+i
        self.__glenable(glLight, val)
        val.bind(glLight)
        
    def LocalViewer(self, val):
        glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, val)

    def MultiSampleAntialias(self, val):
        raise NotImplementedError()

    def MultiSamplemask(self, val):
        raise NotImplementedError()

    def NormalizeNormals(self, val):
        self.__glenable(GL_NORMALIZE, val)

    def PatchSegments(self, val):
        raise NotImplementedError()

    def PointScaleA(self, val):
        raise NotImplementedError()

    def PointScaleB(self, val):
        raise NotImplementedError()

    def PointScaleC(self, val):
        raise NotImplementedError()

    def PointScaleEnable(self, val):
        raise NotImplementedError()

    def PointSize(self, val):
        glPointSize(val)

    def PointSizeMin(self, val):
        raise NotImplementedError()

    def PointSizeMax(self, val):
        raise NotImplementedError()

    def PointSpriteEnable(self, val):
        raise NotImplementedError()

    def RangeFogEnable(self, val):
        raise NotImplementedError()

    def ReadBuffer(self, val):
        glReadBuffer(val)

    def SpecularEnable(self, val):
        raise NotImplementedError()

    def SpecularMaterialSource(self, val):
        raise NotImplementedError()

    def Texture(self, i, tex):
        if self.supportsMultitexture:
            if i < self.noTexUnits:
                glActiveTextureARB(GL_TEXTURE0_ARB + i)
            else:
                raise RuntimeError('Tried to specify non-existant texture unit %s of %s!'
                                   % (i,self.noTexUnits))
        else:
            if i > 0:
                raise RuntimeError('GL_ARB_multitexture is not supported!')
        tex.bind()

    def TweenFactor(self, val):
        raise NotImplementedError()

    def VertexBlend(self, val):
        raise NotImplementedError()

    def WriteBuffer(self, val):
        glDrawBuffer( val )


    ### SPECIAL ImageProcessing OPERATION ###
    ### Where should this go? ###
    def Source(self, val):
        ###should this really be here
        self.Texture(0,val)
        

    def draw2DPlane(self, sizes):
        textures = range(len(sizes))

        glPushAttrib(GL_DEPTH_BUFFER_BIT)
        glDisable(GL_DEPTH_TEST)
        glMatrixMode(GL_PROJECTION)
        glPushMatrix()
        glLoadIdentity()

        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        glLoadIdentity()

        glColor3f(1,1,1)
        glBegin(GL_QUADS)

        for (w,h),i in zip(sizes, textures):
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB+i, 0.0, 0.0)
        glVertex2f(-1,-1)

        for (w,h),i in zip(sizes, textures):
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB+i, w, 0.0)
        glVertex2f(1,-1)

        for (w,h),i in zip(sizes, textures):
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB+i, w, h)
        glVertex2f(1,1)

        for (w,h),i in zip(sizes, textures):
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB+i, 0.0, h)
        glVertex2f(-1,1)

        glEnd()

        glMatrixMode(GL_PROJECTION)
        glPopMatrix()
        glMatrixMode(GL_MODELVIEW)
        glPopMatrix()

        glPopAttrib()


    ##Stuff that can be set on a Target basis
    def Clear(self, val):
        if val:
            glClear(val)


    def ClearColor(self, val):
        glClearColor(*val)


        
class ShaderOpenGLRenderState(OpenGLRenderState):
    
    def __init__(self):
        OpenGLRenderState.__init__(self)

        self.fragmentShader = None
        self.vertexShader = None
        
    def Reset(self):
        if self.vertexShader:
            self.vertexShader.reset()
            self.vertexShader = None

        if self.fragmentShader:
            self.fragmentShader.reset()
            self.fragmentShader = None

        OpenGLRenderState.Reset(self)            
        
    def PixelShader(self, shader):
        self.FragmentShader(shader)

    def FragmentShader(self, shader):
        p = shader.bindFragment(self.resolver)
        self.fragmentShader = p

    def VertexShader(self, shader):
        p = shader.bindVertex(self.resolver)
        self.vertexShader = p


