# 
# Copyright (C) 2003 Carl Johan Lejdfors
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# Please send all bugreports to: Carl Johan Lejdfors <calle.lejdfors@cs.lth.se>
# 

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from constants import *
from PIL import *
import PIL.Image
import sys
import os
import struct

from FX.util import *
import FX

from Numeric import *

GL_TEXTURE_MAX_LEVEL = 0x813D

class OpenGLTexture:
    ## This class contain code borrowed from the tex_object.py in the Cg-tree. 
    ## /Calle Lejdfors

    basePath = ""
    
    def __init__(self,
                 filename = None,
                 type = TEXTURE_2D ):
        self.filename = filename
        self.type = type
        self.texNo = -1

        if type == TEXTURE_RECT:
            self.mipFilter = None
        else:
            self.mipFilter = LINEAR
        self.minFilter = NEAREST
        self.magFilter = LINEAR

        self.initialize()
        
    def maketex(self, target, mipmap = True):
        w,h = self.size
        data = self.data
            
        if mipmap:
            gluBuild2DMipmaps(target, self.format, w, h,
                              self.format, GL_UNSIGNED_BYTE, data)
        else:
            glTexImage2D(target, 0, self.format, w, h,
                         0, self.format, GL_UNSIGNED_BYTE, data)

    def open(self, filename):
        return PIL.Image.open(os.path.join(OpenGLTexture.basePath,filename))

    def initialize(self, mipmap = True):
        self.texNo = glGenTextures(1)
        glBindTexture(self.type, self.texNo)

        if self.type == TEXTURE_1D:
            if self.filename:
                image = self.open(self.filename)
                self.size = image.size
                if image.mode == "RGB":
                    self.format = GL_RGB
                elif image.mode == "RGBA":
                    self.format = GL_RGBA
                else:
                    print "unknown mode!"
                self.data = image.tostring("raw", image.mode, 0, -1)

            data = self.data
            w,h = self.size
            fmt = self.format
            
            if mipmap:
                gluBuild1DMipmaps(self.type, fmt, w,
                                  fmt, GL_UNSIGNED_BYTE, data)
            else:
                glTexImage1D(self.type, 0, GL_RGBA, w, h,
                             0, self.format, GL_UNSIGNED_BYTE, data)

        elif self.type == TEXTURE_2D or self.type == TEXTURE_RECT: 
            if self.filename:
                image = self.open(self.filename)
                self.size = image.size
                if image.mode == "RGB":
                    self.format = GL_RGB
                elif image.mode == "RGBA":
                    self.format = GL_RGBA
                else:
                    print "unknown mode!"
                self.data = image.tostring("raw", image.mode, 0, -1)

            if self.type == TEXTURE_RECT:
                self.maketex(self.type, False)
            else:
                self.maketex(self.type, mipmap)

        elif self.type == TEXTURE_CUBE:
            targets = [ GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
                        GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
                        GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
                        GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
                        GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
                        GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB ]
            for file,target in zip(self.filename,targets):
                image = self.open(file)
                self.size = image.size
                if image.mode == "RGB":
                    self.format = GL_RGB
                elif image.mode == "RGBA":
                    self.format = GL_RGBA
                else:
                    print "unknown mode!"
                self.data = image.tostring("raw", image.mode, 0, -1)
                self.maketex(target,mipmap)
            
        else:
            raise NotImplementedError('Non-1D/2D/CUBE textures are not implemented yet!')

    def textureId(self):
        if self.texNo < 0:
            self.initialize()
        return self.texNo        

    def bind(self, doEnable):
        texNo = self.textureId()
        if doEnable:
            glEnable(self.type)
            glBindTexture(self.type, texNo)

    def texParameter(self, pname, i):
        if type(i) == type(int()):
            glTexParameteri(self.type, pname, i);
        elif type(i) == type(float()):
            glTexParameterf(self.type, pname, i);
        elif type(i) == type([]):
            if len(i) > 0:
                if type(i[0]) == type(int()):
                    glTexParameteriv(self.type, pname, i)
                elif type(i[0]) == type(float()):
                    glTexParameterfv(self.type, pname, i)

    def _setFilters(self):
        if self.mipFilter == NEAREST:
            if self.minFilter == NEAREST:
                self.texParameter(GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST)
            elif self.minFilter == LINEAR:
                self.texParameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST)
            else:
                raise RuntimeError("Unknown MinFilter %x" % self.minFilter)

        elif self.mipFilter == LINEAR:
            if self.minFilter == NEAREST:
                self.texParameter(GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR)
            elif self.minFilter == LINEAR:
                self.texParameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR)
            else:
                raise RuntimeError("Unknown MinFilter %x" % self.minFilter)

        elif self.mipFilter == None:
            if self.minFilter == NEAREST:
                self.texParameter(GL_TEXTURE_MIN_FILTER, GL_NEAREST)
            elif self.minFilter == LINEAR:
                self.texParameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR)

        else:
            raise RuntimeError("Trying to set MinFilter %x with unknown MipFilter %x" % \
                               (self.minFilter,self.mipFilter))
    def MinFilter(self, val):
        self.minFilter = val
        self._setFilters()
            
    def MipFilter(self,val):
        self.mipFilter = val
        self._setFilters()
        
    def MagFilter(self, val):
        self.texParameter(GL_TEXTURE_MAG_FILTER, val)

    def BaseLevel(self, val):
        self.texParameter(GL_TEXTURE_BASE_LEVEL, val)

    def MaxLevel(self, val):
        self.texParameter(GL_TEXTURE_MAX_LEVEL, val)
        
    def WrapS(self, val):
        self.texParameter(GL_TEXTURE_WRAP_S, val)

    def WrapT(self, val):
        self.texParameter(GL_TEXTURE_WRAP_T, val)

    def WrapR(self, val):
        self.texParameter(GL_TEXTURE_WRAP_R, val)

    def TexGenS(self, val):
        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, val)
        glEnable(GL_TEXTURE_GEN_S)

    def TexGenT(self, val):
        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, val)
        glEnable(GL_TEXTURE_GEN_T)

    def TexGenR(self, val):
        glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, val)
        glEnable(GL_TEXTURE_GEN_R)

    def TextureOp(self, val):
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, val)

    def ColorArg(self, i, val):
        glTexEnvi(GL_TEXTURE_ENV, SOURCE0_RGB_NV+i, GL_TEXTURE)

    def ColorOp(self, val):
        self.TextureOp(val)



class DynamicTexture(OpenGLTexture):
    def __init__(self, type = TEXTURE_2D):
        OpenGLTexture.__init__(self,type = type)
        
    def initialize(self):
        w = glutGet(GLUT_WINDOW_WIDTH)
        h = glutGet(GLUT_WINDOW_HEIGHT)

        self.size = w,h

        self.texNo = glGenTextures(1)
        #glBindTexture(self.type, self.texNo)
        #glTexImage2D(self.type, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_FLOAT, 0)
        
    def bind(self, doEnable):
        texNo = self.textureId()

        if doEnable:
            glEnable(self.type)
            glBindTexture(self.type, texNo)
        
        w,h = self.size
        glCopyTexImage2D(self.type, 0, GL_RGB, 0, 0, w, h, 0)
        #glCopyTexSubImage2D(self.type, 0, 0, 0, 0, 0, w, h)
        
        

class PBufferTexture(OpenGLTexture):
    def __init__(self, pbuf,
                 type = TEXTURE_RECT,
                 ):
        self.pbuf = pbuf
        self.type = type
        self.texNo = self.pbuf.textureId()
        self.size = pbuf.GetWidth(),pbuf.GetHeight(),
        

    def textureId(self):
        return self.texNo

    def initialize(self, mipmap = True):
        self.texNo = self.pbuf.textureId()

    def bind(self, doEnable):
        if doEnable:
            glEnable(self.type)
            glBindTexture(self.type, self.texNo)

        #print "Pbuffer active?", self.pbuf.isActive()
        if self.pbuf.isActive():            
            self.pbuf.copyToTexture(doEnable)

            #         w,h = self.size
            #         str = self.pbuf.pixels()
            #         im = Image.fromstring("RGBA", (w,h), str)
            #         im = im.transpose(Image.FLIP_TOP_BOTTOM)
            #         im.save('pbuffer.png')
            

class RawTexture(OpenGLTexture):
    ##!!NOT TESTED!!
    def __init__(self, size, data, format, dataformat = "B"):
        ls = len(size)
        if ls == 1:
            target = GL_TEXTURE_1D
        elif ls == 2:
            target = GL_TEXTURE_2D
        else:
            target = GL_TEXTURE_3D

        OpenGLTexture.__init__(self)
        self.target = target
        self.size = size
        self.data = data

        if format == "RGBA":
            self.format = GL_RGBA
        elif format == "RGB":
            self.format = GL_RBA
        elif format == "D":
            self.format = GL_DEPTH_COMPONENT

        if dataformat == "b":
            self.type = GL_BYTE
        elif dataformat == "B":
            self.type = GL_UNSIGND_BYTE
        elif dataformat == "f":
            self.type = GL_FLOAT
        else:
            self.type = dataformat

    def initialize(self, mipmap):
        self.texNo = glGenTextures(1)
        glBindTexture(self.target, self.texNo)
        if self.target == GL_TEXTURE_1D:
            glTexImage1D(GL_TEXTURE_1D, 0, self.format,
                         size, 0, self.format, self.type, self.data)
        elif self.target == GL_TEXTURE_2D:
            w,h = size
            glTexImage2D(GL_TEXTURE_2D, 0, self.format,
                         w, h, 0, self.format, self.type, self.data)
        else:
            w,h,d = size
            glTexImage3D(GL_TEXTURE_2D, 0, self.format,
                         w, h, h, 0, self.format, self.type, self.data)
            
            
class Sampler:

    def __init__(self, texture, **kwords):
        self.texture = texture
        self.kwords = kwords

    def textureId(self):
        return self.texture.textureId()

    def size(self):
        return self.texture.size

    def bind(self, doEnable = 1):
        self.texture.bind(doEnable)
        for key,val in self.kwords.items():
            try:
                marshalFX(self.texture, key, val)
            except AttributeError, e:
                print "Unknown texture state: '%s': %s" % (key,str(e))

        
